import json 

from msfabricpysdkcore.coreapi import FabricClientCore

class Item:
    """Class to represent a item in Microsoft Fabric"""

    def __init__(self, id, display_name, type, workspace_id, core_client: FabricClientCore, properties = None, definition=None, description="") -> None:
        
        self.id = id
        self.display_name = display_name
        self.description = description
        self.type = type
        self.definition = definition
        self.properties = properties
        self.workspace_id = workspace_id
        
        self.core_client = core_client

    def __str__(self) -> str:
        """Return a string representation of the workspace object"""
        dict_ = {
            'id': self.id,
            'display_name': self.display_name,
            'description': self.description,
            'type': self.type,
            'definition': self.definition,
            'workspace_id': self.workspace_id,
            'properties': self.properties
        }
        return json.dumps(dict_, indent=2)
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def from_dict(item_dict, core_client):
        """Create Item object from dictionary"""
        
        return Item(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
                    properties=item_dict.get('properties', None),
                    definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

    def delete(self, type = None):
        """Delete the workspace item"""

        return self.core_client.delete_item(self.workspace_id, self.id, type=type)
    
    def get_definition(self, type = None, format = None, **kwargs):
        """Get the definition of the item"""
        resp_dict = self.core_client.get_item_definition(self.workspace_id, self.id, type=type, format=format, **kwargs)

        self.definition = resp_dict['definition']
        return resp_dict

    def list_connections(self):
        """List connections of an item in a workspace"""
        return self.core_client.list_item_connections(workspace_id=self.workspace_id, item_id=self.id)

    def update(self, display_name = None, description = None, type = None, return_item=False):
        """Update the item"""

        resp_dict = self.core_client.update_item(workspace_id=self.workspace_id, item_id=self.id,
                                                 display_name=display_name, description=description, type=type,
                                                 return_item=return_item)
        if display_name:
            self.display_name = display_name
        if description:
            self.description = description

        return resp_dict

    def update_definition(self, definition, type = None, **kwargs):
        """Update the item definition"""
        if not type:
            type = self.type
        response = self.core_client.update_item_definition(workspace_id=self.workspace_id, item_id=self.id, 
                                                           definition=definition, type=type, **kwargs)

        self.definition = definition 
        return response
    
    # Tags   
    def apply_tags(self, tags):
        """Apply tags to an item in a workspace"""
        return self.core_client.apply_tags(workspace_id=self.workspace_id, item_id=self.id, tags=tags)
    
    def unapply_tags(self, tags):
        """Unapply tags from an item in a workspace"""
        return self.core_client.unapply_tags(workspace_id=self.workspace_id, item_id=self.id, tags=tags)
    
    # Shortcut

    def create_shortcut(self, path, name, target):
        """Create a shortcut in the item"""
        return self.core_client.create_shortcut(workspace_id=self.workspace_id, item_id=self.id,
                                                path=path, name=name, target=target)
    
    def delete_shortcut(self, path, name):
        """Delete the shortcut in the item"""
        return self.core_client.delete_shortcut(workspace_id=self.workspace_id, item_id=self.id,
                                                path=path, name=name)

    def get_shortcut(self, path, name):
        """Get the shortcut in the item"""
        return self.core_client.get_shortcut(workspace_id=self.workspace_id, item_id=self.id,
                                                path=path, name=name)
    
    def list_shortcuts(self, parent_path = None):
        return self.core_client.list_shortcuts(workspace_id=self.workspace_id, item_id=self.id, parent_path=parent_path)

    # Job Scheduler

    def cancel_item_job_instance(self, job_instance_id):
        """Cancel a job instance ofjob the item"""
        return self.core_client.cancel_item_job_instance(workspace_id=self.workspace_id, item_id=self.id,
                                                         job_instance_id=job_instance_id)
    
    def create_item_schedule(self, job_type, configuration, enabled):
        """Create a schedule for the job instance"""
        return self.core_client.create_item_schedule(workspace_id=self.workspace_id,
                                                     item_id=self.id,
                                                     job_type=job_type,
                                                     configuration=configuration,
                                                     enabled=enabled)
    
    def get_item_job_instance(self, job_instance_id):
        """Get the job instance of the item"""
        return self.core_client.get_item_job_instance(workspace_id=self.workspace_id, item_id=self.id,
                                                      job_instance_id=job_instance_id)
    def get_item_schedule(self, job_type, schedule_id):
        """Get the schedule for the job instance"""
        return self.core_client.get_item_schedule(workspace_id=self.workspace_id,
                                                 item_id=self.id,
                                                 job_type=job_type,
                                                 schedule_id=schedule_id)
    
    def list_item_job_instances(self):
        """List all job instances for the job instance"""
        return self.core_client.list_item_job_instances(workspace_id=self.workspace_id,
                                                        item_id=self.id)                                      
    
    def list_item_schedules(self, job_type):
        """List all schedules for the job instance"""
        return self.core_client.list_item_schedules(workspace_id=self.workspace_id,
                                                  item_id=self.id,
                                                  job_type=job_type)

    def run_on_demand_item_job(self, job_type, execution_data = None):
        return self.core_client.run_on_demand_item_job(workspace_id=self.workspace_id, item_id=self.id,
                                                       job_type=job_type, execution_data=execution_data)
    
    def update_item_schedule(self, job_type, schedule_id, configuration, enabled):
        """Update the schedule for the job instance"""
        return self.core_client.update_item_schedule(workspace_id=self.workspace_id,
                                                     item_id=self.id,
                                                     job_type=job_type,
                                                     schedule_id=schedule_id,
                                                     configuration=configuration,
                                                     enabled=enabled)

    
    # External Data Shares

    def create_external_data_share(self, paths, recipient):
        return self.core_client.create_external_data_share(workspace_id=self.workspace_id, item_id=self.id,
                                                           paths=paths, recipient=recipient)

    def get_external_data_share(self, external_data_share_id):
        return self.core_client.get_external_data_share(workspace_id=self.workspace_id, item_id=self.id,
                                                       external_data_share_id=external_data_share_id)

    def list_external_data_shares_in_item(self):
        return self.core_client.list_external_data_shares_in_item(workspace_id=self.workspace_id, item_id=self.id)
    
    def revoke_external_data_share(self, external_data_share_id):
        return self.core_client.revoke_external_data_share(workspace_id=self.workspace_id, item_id=self.id,
                                                          external_data_share_id=external_data_share_id)

    # One Lake data access security

    def list_data_access_roles(self):
        return self.core_client.list_data_access_roles(workspace_id=self.workspace_id, item_id=self.id)
    
    def create_or_update_data_access_roles(self, data_access_roles, dryrun = False, etag_match = None):
        return self.core_client.create_or_update_data_access_roles(workspace_id=self.workspace_id, item_id=self.id,
                                                                     data_access_roles=data_access_roles, dryrun=dryrun, etag_match=etag_match)


from msfabricpysdkcore.item import Item

class Environment(Item):
    """Class to represent a item in Microsoft Fabric"""

    def __init__(self, id, display_name, type, workspace_id, auth, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, auth, properties, definition, description)

    def from_dict(item_dict, core_client):
        return Environment(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""),
            auth=core_client)

    def get_published_settings(self):
        """Get the published settings of the environment"""
        return self.core_client.get_published_settings(self.workspace_id, self.id)
    
    def get_staging_settings(self):
        """Get the staging settings of the environment"""
        return self.core_client.get_staging_settings(self.workspace_id, self.id)


    def update_staging_settings(self,
                                driver_cores = None, driver_memory = None, dynamic_executor_allocation = None,
                                executor_cores = None, executor_memory = None, instance_pool = None,
                                runtime_version = None, spark_properties = None):
        """Update the staging settings of the environment"""
        return self.core_client.update_staging_settings(self.workspace_id, self.id, driver_cores, driver_memory,
                                                        dynamic_executor_allocation, executor_cores, executor_memory,
                                                        instance_pool, runtime_version, spark_properties)

    def get_published_libraries(self):
        """Get the published libraries of the environment"""
        return self.core_client.get_published_libraries(self.workspace_id, self.id)
    
    def get_staging_libraries(self):
        """Get the staging libraries of the environment"""
        return self.core_client.get_staging_libraries(self.workspace_id, self.id)
    
    def upload_staging_library(self, file_path):
        return self.core_client.upload_staging_library(self.workspace_id, self.id, file_path)
    
    def delete_staging_library(self, library_to_delete):
        return self.core_client.delete_staging_library(self.workspace_id, self.id, library_to_delete)
    
    def publish_environment(self):
        return self.core_client.publish_environment(self.workspace_id, self.id) 

    def cancel_publish(self):
        return self.core_client.cancel_publish(self.workspace_id, self.id)


from msfabricpysdkcore.item import Item

class Lakehouse(Item):
    """Class to represent a item in Microsoft Fabric"""

    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return Lakehouse(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

    def list_tables(self):
        """List all tables in the lakehouse"""
        return self.core_client.list_tables(self.workspace_id, self.id)
    
    def load_table(self, table_name, path_type, relative_path,
                    file_extension = None, format_options = None,
                    mode = None, recursive = None, wait_for_completion = True):
        """Load a table in the lakehouse"""
        return self.core_client.load_table(self.workspace_id, self.id, table_name, path_type, relative_path,
                                           file_extension, format_options, mode, recursive, wait_for_completion)
    
    def run_on_demand_table_maintenance(self, execution_data, job_type = "TableMaintenance", wait_for_completion = True):
        """Run on demand table maintenance"""
        return self.core_client.run_on_demand_table_maintenance(self.workspace_id, self.id, execution_data, job_type, wait_for_completion)
    
    def list_livy_sessions(self):
        """List all livy sessions in the lakehouse"""
        return self.core_client.list_lakehouse_livy_sessions(self.workspace_id, self.id)

    def get_livy_session(self, livy_id):
        """Get a livy session in the lakehouse"""
        return self.core_client.get_lakehouse_livy_session(self.workspace_id, self.id, livy_id)


from .coreapi import FabricClientCore
from .adminapi import FabricClientAdmin
from .fabric_azure_client import FabricAzureClient

__all__ = ["FabricClientCore", "FabricClientAdmin", "FabricAzureClient"]



import logging
from abc import abstractmethod
import os
from warnings import warn
from time import sleep
import requests
import json

from msfabricpysdkcore.auth import FabricAuthClient, FabricServicePrincipal, FabricSparkUtilsAuthentication, MSALConfidentialClientApplicationAuthentication
from msfabricpysdkcore.util import logger

class FabricClient():
    """FabricClient class to interact with Fabric API"""

    _logger: logging.Logger

    def __init__(self, scope, tenant_id = None, client_id = None, client_secret = None, username = None, password = None, silent=None) -> None:
        """Initialize FabricClient object"""

        self._logger = logger.getChild(__name__)

        self.tenant_id = tenant_id if tenant_id else os.getenv("FABRIC_TENANT_ID")
        self.client_id = client_id if client_id else os.getenv("FABRIC_CLIENT_ID")
        self.client_secret = client_secret if client_secret else os.getenv("FABRIC_CLIENT_SECRET")
        self.username = username if username else os.getenv("FABRIC_USERNAME")
        self.password = password if password else os.getenv("FABRIC_PASSWORD")
        self.scope = scope

        if self.client_id is None or self.client_secret is None or self.tenant_id is None:
            try:
                self.auth = FabricSparkUtilsAuthentication(self.scope)
            except:
                self.auth = FabricAuthClient(self.scope)
        else:
            if username and password:
                self.auth = MSALConfidentialClientApplicationAuthentication(tenant_id = self.tenant_id,
                                                                            client_id = self.client_id,
                                                                            client_secret = self.client_secret,
                                                                            username = self.username,
                                                                            password = self.password,
                                                                            scope = self.scope)
            else:
                self.auth = FabricServicePrincipal(scope= self.scope,
                                                   tenant_id = self.tenant_id,
                                                   client_id = self.client_id, 
                                                   client_secret = self.client_secret)

        if silent is not None:
            warn("The 'silent' parameter is deprecated and will be removed in a future version.", DeprecationWarning, stacklevel=2)

            
    def get_token(self):
        """Get token from Entra"""
        return self.auth.get_token()

    def calling_routine(self, url, operation, body = None, headers=None, file_path = None, response_codes = [200], error_message = "Error",
                        continue_on_error_code = False, return_format = "value_json", paging = False,
                        wait_for_completion = True, continuation_token = None):
        """Routine to make API calls
        Args:
            url (str): The URL of the API
            operation (str): The operation to perform
            body (dict): The body of the request
            response_codes (list): The response codes to expect
            error_message (str): The error message
            continue_on_error_code (bool): Whether to continue on error code
            return_format (str): The format of the return
            paging (bool): Whether to paginate
            wait_for_completion (bool): Whether to wait for the operation to complete
        Returns:
            dict: The response
        """
        original_url = url

        if continuation_token:
            last_part_url = url.split("/")[-1]   
            if "?" not in last_part_url:
                continuation_token  = f"?continuationToken={continuation_token}"
            else:
                continuation_token  = f"&continuationToken={continuation_token}"
            url = f"{url}{continuation_token}"

        response_codes.append(429)
        if headers is None:
            headers = self.auth.get_headers()
        for _ in range(10):
            if operation == "GET":
                response = requests.get(url=url, headers=headers)
            elif operation == "PATCH":
                if body is None:
                    response = requests.patch(url=url, headers=headers)
                else:
                    response = requests.patch(url=url, headers=headers, json=body)
            elif operation == "POST":
                if body is not None:
                    response = requests.post(url=url, headers=headers, json=body)
                elif file_path is not None:
                    headers.pop('Content-Type')
                    with open(file_path, 'rb') as f:
                        files = {"file": f}
                        response = requests.post(url=url, files=files, headers=headers)
                else:
                    response = requests.post(url=url, headers=headers)
            elif operation == "PUT":
                if body is None:
                    response = requests.put(url=url, headers=headers)
                else:
                    response = requests.put(url=url, headers=headers, json=body)
            elif operation == "DELETE":
                response = requests.delete(url=url, headers=headers)
            else:
                raise ValueError("Invalid operation")
            if response.status_code == 429:
                self._logger.info("Too many requests, waiting 10 seconds")
                sleep(10)
                continue
            elif response.status_code == 202:
                if wait_for_completion:
                    operation_result = self.long_running_operation(response.headers)
                    if "operation_result" in return_format:
                        return operation_result
                return response
            elif response.status_code not in response_codes:
                if continue_on_error_code:
                    return response
                raise Exception(f"{error_message}: {response.status_code} {response.text}")
            break

        if paging:
            resp_dict = json.loads(response.text)

            if return_format in ["data", "itemEntities", "Overrides", "accessEntities", "workspaces"]:
                items = resp_dict[return_format]
            else:
                items = resp_dict["value"]

                
            if "continuationToken" in resp_dict and resp_dict["continuationToken"]:
                continuation_token = resp_dict["continuationToken"]
                items_next = self.calling_routine(url=original_url, operation=operation, body=body, headers=headers,
                                                  response_codes=response_codes,
                                                  error_message=error_message, continuation_token=continuation_token,
                                                  return_format=return_format, paging=True, wait_for_completion=wait_for_completion)
                items.extend(items_next)
            if "etag" in return_format:
                return items, response.headers.get('ETag')
            return items

        if "value_json" in return_format:
            resp_dict = json.loads(response.text)
            if "etag" in return_format:
                return resp_dict["value"], response.headers.get('ETag')	
            return resp_dict["value"]
        
        if "json" in return_format:
            return json.loads(response.text)

        return response
    
    @abstractmethod
    def long_running_operation(self, headers):
        """Long running operation"""
        pass


import json
from msfabricpysdkcore.coreapi import FabricClientCore


class SparkCustomPool:
    """Class to represent a custom pool in Microsoft Fabric"""

    def __init__(self, id, name, type, node_family, node_size, auto_scale, dynamic_executor_allocation, workspace_id, core_client: FabricClientCore) -> None:
        
        self.id = id
        self.name = name
        self.type = type
        self.node_family = node_family
        self.node_size = node_size
        self.auto_scale = auto_scale
        self.dynamic_executor_allocation = dynamic_executor_allocation
        self.workspace_id = workspace_id
        
        self.core_client = core_client

    def __str__(self) -> str:
        """Return a string representation of the workspace object"""
        dict_ = {
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "nodeFamily": self.node_family,
            "nodeSize": self.node_size,
            "autoScale": self.auto_scale,
            "dynamicExecutorAllocation": self.dynamic_executor_allocation,
            "workspaceId": self.workspace_id
        }
        return json.dumps(dict_, indent=2)
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def from_dict(item_dict, core_client):
        """Create Item object from dictionary"""

        if 'autoScale' not in item_dict:
            item_dict['autoScale'] = item_dict['auto_scale']

        if 'dynamicExecutorAllocation' not in item_dict:
            item_dict['dynamicExecutorAllocation'] = item_dict['dynamic_executor_allocation']
        
        if 'nodeFamily' not in item_dict:
            item_dict['nodeFamily'] = item_dict['node_family']
        
        if 'nodeSize' not in item_dict:
            item_dict['nodeSize'] = item_dict['node_size']
        
        return SparkCustomPool(id=item_dict['id'], name=item_dict['name'], type=item_dict['type'], node_family=item_dict['nodeFamily'],
                                node_size=item_dict['nodeSize'], auto_scale=item_dict['autoScale'], dynamic_executor_allocation=item_dict['dynamicExecutorAllocation'],
                                workspace_id=item_dict['workspaceId'], core_client=core_client)


    def delete(self):
        """Delete the custom pool item"""
        return self.core_client.delete_workspace_custom_pool(self.workspace_id, self.id)
    

    def update(self, name, node_family, node_size, auto_scale, dynamic_executor_allocation):
        """Update the custom pool item"""
        _ = self.core_client.update_workspace_custom_pool(self.workspace_id, self.id, name, node_family, node_size, auto_scale, dynamic_executor_allocation)
        if name is not None:
            self.name = name
        if node_family is not None:
            self.node_family = node_family
        if node_size is not None:
            self.node_size = node_size
        if auto_scale is not None:
            self.auto_scale = auto_scale
        if dynamic_executor_allocation is not None:
            self.dynamic_executor_allocation = dynamic_executor_allocation

        return self
    


from msfabricpysdkcore.item import Item

class Eventstream(Item):
    """Class to represent a eventstream in Microsoft Fabric"""

    def __init__(self, id, display_name, type, workspace_id, core_client, properties=None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)
    
    def from_dict(item_dict, core_client):
        return Eventstream(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

    def get_definition(self, type=None, format=None):
        """Method to get the definition of the eventstream"""
        return super().get_definition(type="eventstreams", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the eventstream"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="eventstreams")
    
    # eventstream topology
    def get_eventstream_destination(self, destination_id):
        """Get the destination of an eventstream in a workspace"""
        return self.core_client.get_eventstream_destination(workspace_id=self.workspace_id, eventstream_id=self.id, destination_id=destination_id)
    
    def get_eventstream_destination_connection(self, destination_id):
        """Get the connection of a destination in an eventstream in a workspace"""
        return self.core_client.get_eventstream_destination_connection(workspace_id=self.workspace_id, eventstream_id=self.id, destination_id=destination_id)

    def get_eventstream_source(self, source_id):
        """Get the source of an eventstream in a workspace"""
        return self.core_client.get_eventstream_source(workspace_id=self.workspace_id, eventstream_id=self.id, source_id=source_id)

    def get_eventstream_source_connection(self, source_id):
        """Get the connection of a source in an eventstream in a workspace"""
        return self.core_client.get_eventstream_source_connection(workspace_id=self.workspace_id, eventstream_id=self.id, source_id=source_id)

    def get_eventstream_topology(self):
        """Get the topology of an eventstream in a workspace"""
        return self.core_client.get_eventstream_topology(workspace_id=self.workspace_id, eventstream_id=self.id)
    
    def pause_eventstream(self):
        """Pause an eventstream in a workspace"""
        return self.core_client.pause_eventstream(workspace_id=self.workspace_id, eventstream_id=self.id)
    
    def pause_eventstream_destination(self, destination_id):
        """Pause a destination in an eventstream in a workspace"""
        return self.core_client.pause_eventstream_destination(workspace_id=self.workspace_id, eventstream_id=self.id, destination_id=destination_id)

    def pause_eventstream_source(self, source_id):
        """Pause a source in an eventstream in a workspace"""
        return self.core_client.pause_eventstream_source(workspace_id=self.workspace_id, eventstream_id=self.id, source_id=source_id)

    def resume_eventstream(self, start_type, custom_start_date_time = None):
        """Resume an eventstream in a workspace"""
        return self.core_client.resume_eventstream(workspace_id=self.workspace_id, eventstream_id=self.id, start_type=start_type, custom_start_date_time=custom_start_date_time)

    def resume_eventstream_destination(self, destination_id, start_type, custom_start_date_time = None):
        """Resume a destination in an eventstream in a workspace"""
        return self.core_client.resume_eventstream_destination(workspace_id=self.workspace_id, eventstream_id=self.id,
                                                               destination_id=destination_id, start_type=start_type, custom_start_date_time=custom_start_date_time)

    def resume_eventstream_source(self, source_id, start_type, custom_start_date_time = None):
        """Resume a source in an eventstream in a workspace"""
        return self.core_client.resume_eventstream_source(workspace_id=self.workspace_id, eventstream_id=self.id,
                                                          source_id=source_id, start_type=start_type, custom_start_date_time=custom_start_date_time)




import requests
import json
import os
from time import sleep
from warnings import warn

from msfabricpysdkcore.client import FabricClient

class FabricClientAdmin(FabricClient):
    """FabricClientAdmin class to interact with Fabric Admin APIs"""

    def __init__(self, tenant_id = None, client_id = None, client_secret = None,
                 username = None, password = None) -> None:
        """Initialize FabricClientAdmin object"""
        super().__init__(scope="https://api.fabric.microsoft.com/.default", 
                         tenant_id=tenant_id, client_id=client_id, client_secret=client_secret,
                         username=username, password=password)


    def long_running_operation(self, response_headers):
        """Check the status of a long running operation"""
        from msfabricpysdkcore.coreapi import FabricClientCore
        fc = FabricClientCore(tenant_id=self.tenant_id, client_id=self.client_id, client_secret=self.client_secret)

        return fc.long_running_operation(response_headers)

    # Domain APIs

    def assign_domain_workspaces_by_ids(self, domain_id, workspaces_ids):
        """Assign workspaces to a domain by workspace IDs
        
        Args:
            domain_id (str): The ID of the domain
            workspace_ids (list): The list of workspace IDs
            wait_for_completion (bool): Whether to wait for the operation to complete
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/assignWorkspaces"
        body = {
            "workspacesIds": workspaces_ids
        }

        response:requests.Response = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                          error_message = "Error assigning workspaces by ids", return_format="response")

        return response.status_code
    
    def assign_domains_workspaces_by_principals(self, domain_id, principals, wait_for_completion=True):
        """Assign workspaces to a domain by principals
        
        Args:
            domain_id (str): The ID of the domain
            principals (list): The list of principal IDs
            wait_for_completion (bool): Whether to wait for the operation to complete
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/assignWorkspacesByPrincipals"
        body = {
            "principals": principals
        }

        response:requests.Response = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [202, 429],
                                                          error_message = "Error assigning workspaces by principals", wait_for_completion=wait_for_completion,
                                                          return_format="response")

        return response.status_code
    
    def create_domain(self, display_name, description = None, parent_domain_id = None):
        """Method to create a domain
        
        Args:
            display_name (str): The display name of the domain
            description (str): The description of the domain
            parent_domain_id (str): The parent domain ID
        Returns:
            Domain: The Domain object
        """
        from msfabricpysdkcore.domain import Domain

        url = "https://api.fabric.microsoft.com/v1/admin/domains"
        body = {
            "displayName": display_name
        }
        if description:
            body["description"] = description
        if parent_domain_id:
            body["parentDomainId"] = parent_domain_id

        domain_dict = self.calling_routine(url = url, operation = "POST", body = body, return_format="json",
                                           response_codes = [200, 201, 429], error_message = "Error creating domain")
    
        domain = Domain.from_dict(domain_dict, self)
        return domain
    
    def delete_domain(self, domain_id):
        """Method to delete a domain
        
        Args:
            domain_id (str): The ID of the domain
        Returns:
            status_code (int): The status code of the request
        """

        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}"

        response:requests.Response = self.calling_routine(url = url, operation = "DELETE", response_codes = [200, 429],
                                                          error_message = "Error deleting domain",
                                                          return_format="response")

        return response.status_code
    
    def get_domain_by_id(self, domain_id):
        """Method to get a domain by ID
        
        Args:
            domain_id (str): The ID of the domain
        Returns:
            Domain: The Domain object
        """
        from msfabricpysdkcore.domain import Domain

        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}"

        domain_dict = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429], error_message = "Error getting domain",
                                           return_format="json")
        domain = Domain.from_dict(domain_dict, self)
        return domain

    def get_domain_by_name(self, domain_name):
        """Method to get a domain by name
        
        Args:
            domain_name (str): The name of the domain
        Returns:
            Domain: The Domain object
        """
        domains = self.list_domains()
        domains = [domain for domain in domains if domain.display_name == domain_name]
        if len(domains) > 0:
            return self.get_domain_by_id(domains[0].id)
        raise ValueError("Domain not found")
    
    def get_domain(self, domain_id = None, domain_name = None):
        """Get a domain by ID or name
        
        Args:
            domain_id (str): The ID of the domain
            domain_name (str): The name of the domain
        Returns:
            Domain: The Domain object
        """
        if domain_id:
            return self.get_domain_by_id(domain_id)
        if domain_name:
            return self.get_domain_by_name(domain_name)
        raise ValueError("Either domain_id or domain_name must be provided")
    
    def list_domain_workspaces(self, domain_id, workspace_objects = False):
        """List workspaces in a domain
        
        Args:
            domain_id (str): The ID of the domain
        Returns:
            list: List of Workspace objects
        """

       
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/workspaces"

        workspaces = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429], error_message = "Error listing domain workspaces",
                                          return_format="value_json", paging=True)

        if workspace_objects:
            from msfabricpysdkcore import FabricClientCore
            fc = FabricClientCore(tenant_id=self.tenant_id, client_id=self.client_id, client_secret=self.client_secret)
            workspaces = [fc.get_workspace_by_id(workspace["id"]) for workspace in workspaces]

        return workspaces
    
    def list_domains(self, nonEmptyOnly=False):
        """List all domains in the tenant
        
        Args:
            nonEmptyOnly (bool): Whether to list only non-empty domains
        Returns:
            list: List of Domain objects"""
        from msfabricpysdkcore.domain import Domain

        url = "https://api.fabric.microsoft.com/v1/admin/domains"
        if nonEmptyOnly:
            url = f"{url}?nonEmptyOnly=True"

        resp_dict = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429], error_message = "Error listing domains",
                                         return_format="json")
        domains = resp_dict["domains"]

        domains = [Domain.from_dict(i, self) for i in domains]
        return domains

    def role_assignments_bulk_assign(self, domain_id, type, principals):
        """Assign a role to principals in bulk
        
        Args:
            domain_id (str): The ID of the domain
            type (str): The type of the role assignment
            principals (list): The list of principals
        Returns:
            int: The status code of the response
        """

        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/roleAssignments/bulkAssign"
        body = {
            "type": type,
            "principals": principals
        }

        response: requests.Response = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                           error_message = "Error bulk assigning role assignments",
                                                           return_format="response")

        return response.status_code
    
    def role_assignments_bulk_unassign(self, domain_id, type, principals):
        """Unassign a role from principals in bulk
        
        Args:
            domain_id (str): The ID of the domain
            type (str): The type of the role assignment
            principals (list): The list of principals
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/roleAssignments/bulkUnassign"
        body = {
            "type": type,
            "principals": principals
        }

        response: requests.Response = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                           error_message = "Error bulk unassigning role assignments",
                                                           return_format="response")

        return response.status_code

    def unassign_all_domain_workspaces(self, domain_id):
        """Unassign all workspaces from a domain
        
        Args:
            domain_id (str): The ID of the domain
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/unassignAllWorkspaces"

        response:requests.Response = self.calling_routine(url = url, operation = "POST", response_codes = [200, 429],
                                                          error_message = "Error unassigning all workspaces",
                                                          return_format="response")
        return response.status_code
    
    def unassign_domain_workspaces_by_ids(self, domain_id, workspace_ids):
        """Unassign workspaces from a domain by workspace IDs
        
        Args:
            domain_id (str): The ID of the domain
            workspace_ids (list): The list of workspace IDs
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/unassignWorkspaces"
        body = {
            "workspacesIds": workspace_ids
        }

        response:requests.Response = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                          error_message = "Error unassigning workspaces by ids",
                                                          return_format="response")

        return response.status_code
    
    def update_domain(self, domain_id, description = None, display_name = None, contributors_scope = None, return_item = False):
        """Method to update a domain
        
        Args:
            domain_id (str): The ID of the domain
        Returns:
            Domain: The Domain object
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}"
        body = {}
        if description:
            body["description"] = description

        if display_name:
            body["displayName"] = display_name

        if contributors_scope:
            body["contributorsScope"] = contributors_scope

        response_json = self.calling_routine(url = url, operation = "PATCH", body = body,
                                             response_codes = [200, 429], error_message = "Error updating domain",
                                             return_format="json")

        if return_item:
            return self.get_domain_by_id(domain_id)
        return response_json

    
    def assign_domain_workspaces_by_capacities(self, domain_id, capacities_ids, wait_for_completion=True):
        """Assign workspaces to a domain by capacities
        
        Args:
            domain_id (str): The ID of the domain
            capacities (list): The list of capacity IDs
            wait_for_completion (bool): Whether to wait for the operation to complete
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/domains/{domain_id}/assignWorkspacesByCapacities"

        body = {
            "capacitiesIds": capacities_ids
        }

        response:requests.Response = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [202, 429],
                                                          error_message = "Error assigning workspaces by capacities", wait_for_completion=wait_for_completion,
                                                          return_format="response")

        return response.status_code
    
    # External Data Share APIs
    
    def list_external_data_shares(self):
        # GET https://api.fabric.microsoft.com/v1/admin/items/externalDataShares
        """List external data shares
        
        Returns:
            list: The list of external data shares
        """
        url = "https://api.fabric.microsoft.com/v1/admin/items/externalDataShares"


        items: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                           error_message = "Error listing external data shares",
                                           return_format="value_json", paging=True)

        return items
    
    def revoke_external_data_share(self, external_data_share_id, item_id, workspace_id):
        """Revoke an external data share
        Args:
            external_data_share_id (str): The ID of the external data share
            item_id (str): The ID of the item
            workspace_id (str): The ID of the workspace
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/workspaces/{workspace_id}/items/{item_id}/externalDataShares/{external_data_share_id}/revoke"

        response:requests.Response = self.calling_routine(url = url, operation = "POST", response_codes = [200, 429],
                                                          error_message = "Error revoking external data share",
                                                          return_format="response")

        return response.status_code

    # Items APIs
    
    def get_item(self, item_id, workspace_id, type = None):
        """Get an item from the workspace
        
        Args:
            item_id (str): The ID of the item
            workspace_id (str): The ID of the workspace
            type (str): The type of the item
        Returns:
            AdminItem: The item object
        """
        from msfabricpysdkcore.admin_item import AdminItem

        url = f"https://api.fabric.microsoft.com/v1/admin/workspaces/{workspace_id}/items/{item_id}"
        if type:
            url += f"?type={type}"

        response_json: dict = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                                   error_message = "Error getting item", return_format="json")

        return AdminItem.from_dict(response_json, self)
    
    
    def list_item_access_details(self, workspace_id, item_id, type=None):
        """Get the access details of the item
        
        Args:
            workspace_id (str): The ID of the workspace
            item_id (str): The ID of the item
            type (str): The type of the item
        Returns:
            dict: The access details of the item
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/workspaces/{workspace_id}/items/{item_id}/users"

        if type:
            url += f"?type={type}"

        response_json: dict = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                                   error_message = "Error getting item access details", return_format="json")
           
        return response_json
    
    def list_items(self, workspace_id = None, capacity_id = None, type=None,
                   state=None):
        """List all items

        Returns:
            list: The list of items in the workspace
        """
        from msfabricpysdkcore.admin_item import AdminItem

        url = f"https://api.fabric.microsoft.com/v1/admin/items"
        first_parameter = False
        if workspace_id:
            url = f"{url}?workspaceId={workspace_id}"
            first_parameter = True
        if capacity_id:
            if first_parameter:
                url = f"{url}&capacityId={capacity_id}"
            else:
                url = f"{url}?capacityId={capacity_id}"
                first_parameter = True
        if type:
            if first_parameter:
                url = f"{url}&type={type}"
            else:
                url = f"{url}?type={type}"
                first_parameter = True
        if state:
            if first_parameter:
                url = f"{url}&state={state}"
            else:
                url = f"{url}?state={state}"

        items = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                     error_message = "Error listing items", return_format="itemEntities", paging=True)

        items =  [AdminItem.from_dict(item, self) for item in items]
        return items
    
    # Labels APIs

      
    def bulk_remove_labels(self, items):
        """Remove labels in bulk
        Args:
            items (list): The list of item IDs  
        Returns:
            dict: The response from the API"""

        url = "https://api.fabric.microsoft.com/v1/admin/items/bulkRemoveLabels"

        if len(items) > 2000:
            self.bulk_remove_labels(items[2000:])
            items = items[:2000]
        
        body = {
            "items": items
        }

        response_json: dict = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                   error_message = "Error removing labels", return_format="json")

        return response_json
      
    def bulk_set_labels(self, items, label_id, assignment_method = None, delegated_principal = None):
        """Set labels in bulk
        Args:
            items (list): The list of item IDs
            label_id (str): The ID of the label
            assignment_method (str): The assignment method
            delegated_principal (str): The delegated principal
        Returns:
            dict: The response from the API
        """

        url = "https://api.fabric.microsoft.com/v1/admin/items/bulkSetLabels"

        if len(items) > 2000:
            self.bulk_set_labels(items[2000:], label_id, assignment_method, delegated_principal)
            items = items[:2000]

        body = {
            "items": items,
            "labelId": label_id
        }
        if assignment_method:
            body["assignmentMethod"] = assignment_method

        if delegated_principal:
            body["delegatedPrincipal"] = delegated_principal

        response_json: dict = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                   error_message = "Error setting labels", return_format="json")

        return response_json
    
    # POST https://api.fabric.microsoft.com/v1/admin/items/bulkRemoveSharingLinks
    def bulk_remove_sharing_links(self, items, sharing_link_type):
        """Remove sharing links in bulk
        Args:
            items (list): The list of item IDs
            sharing_link_type (str): The type of the sharing link
        Returns:
            Response: The response from the API
        """
        url = "https://api.fabric.microsoft.com/v1/admin/items/bulkRemoveSharingLinks"

        if len(items) > 500:
            self.bulk_remove_sharing_links(items = items[500:], sharing_link_type = sharing_link_type)
            items = items[:500]

        body = {
            "items": items,
            "sharingLinkType": sharing_link_type
        }

        response = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 202, 429],
                                                   error_message = "Error removing sharing links", return_format="response")

        return response
    
    # POST https://api.fabric.microsoft.com/v1/admin/items/removeAllSharingLinks
    def remove_all_sharing_links(self, sharing_link_type):
        """Remove all sharing links
        Args:
            sharing_link_type (str): The type of the sharing link
        Returns:
            Response: The response from the API
        """
        url = "https://api.fabric.microsoft.com/v1/admin/items/removeAllSharingLinks"

        body = {
            "sharingLinkType": sharing_link_type
        }

        response= self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 202, 429],
                                                   error_message = "Error removing all sharing links", return_format="response")

        return response
    
    # Tags APIs

    #POST https://api.fabric.microsoft.com/v1/admin/tags/bulkCreateTags
    def bulk_create_tags(self, create_tags_request):
        """Create tags in bulk
        Args:
            create_tags_request (list): The request body
        Returns:
            dict: The response from the API
        """
        url = "https://api.fabric.microsoft.com/v1/admin/tags/bulkCreateTags"

        body = {
            "createTagsRequest": create_tags_request
        }

        response_json: dict = self.calling_routine(url = url, operation = "POST", body = body,
                                                   response_codes = [201, 429], error_message = "Error creating tags",
                                                   return_format="json")

        return response_json
    
    # DELETE https://api.fabric.microsoft.com/v1/admin/tags/{tagId}
    def delete_tag(self, tag_id):
        """Delete a tag
        Args:
            tag_id (str): The ID of the tag
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/tags/{tag_id}"

        response:requests.Response = self.calling_routine(url = url, operation = "DELETE", response_codes = [200, 429],
                                                          error_message = "Error deleting tag", return_format="response")

        return response.status_code
    
    # GET https://api.fabric.microsoft.com/v1/admin/tags
    def list_tags(self):
        """List all tags
        Returns:
            list: The list of tags
        """
        url = "https://api.fabric.microsoft.com/v1/admin/tags"

        items: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                           error_message = "Error listing tags", return_format="value_json", paging=True)

        return items

    # PATCH https://api.fabric.microsoft.com/v1/admin/tags/{tagId}
    def update_tag(self, tag_id, display_name = None):
        """Update a tag
        Args:
            tag_id (str): The ID of the tag
            display_name (str): The display name of the tag
        Returns:
            dict: The response from the API
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/tags/{tag_id}"
        body = {}
        if display_name:
            body["displayName"] = display_name

        response_json: dict = self.calling_routine(url = url, operation = "PATCH", body = body, response_codes = [200, 429],
                                                   error_message = "Error updating tag", return_format="json")

        return response_json

    
    # Tenant Settings APIs

    # DELETE https://api.fabric.microsoft.com/v1/admin/capacities/{capacityId}/delegatedTenantSettingOverrides/{tenantSettingName}
    def delete_capacity_tenant_setting_override(self, capacity_id, tenant_setting_name):
        """Delete a tenant setting override at the capacity
        
        Args:
            capacity_id (str): The ID of the capacity
            tenant_setting_name (str): The name of the tenant setting
            
        Returns:
            int: The status code of the response
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/capacities/{capacity_id}/delegatedTenantSettingOverrides/{tenant_setting_name}"

        response:requests.Response = self.calling_routine(url = url, operation = "DELETE", response_codes = [200, 429],
                                                          error_message = "Error deleting capacity tenant setting override",
                                                          return_format="response")

        return response.status_code

    def list_capacities_tenant_settings_overrides(self):
        """Returns list of tenant setting overrides that override at the capacities
        
        Returns:
            list: The capacities tenant settings overrides
        """
        url = "https://api.fabric.microsoft.com/v1/admin/capacities/delegatedTenantSettingOverrides"


        items: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                           error_message = "Error listing capacities tenant settings overrides",
                                           return_format="value", paging=True)

        return items
    
    # GET https://api.fabric.microsoft.com/v1/admin/capacities/{capacityId}/delegatedTenantSettingOverrides?continuationToken={continuationToken}
    def list_capacity_tenant_settings_overrides_by_capacity_id(self, capacity_id):
        """Returns list of tenant setting overrides that override at the capacity
        
        Args:
            capacity_id (str): The ID of the capacity
        Returns:
            list: The capacities tenant settings overrides
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/capacities/{capacity_id}/delegatedTenantSettingOverrides"

        items: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                           error_message = "Error listing capacity tenant settings overrides",
                                           return_format="value", paging=True)

        return items

    # GET https://api.fabric.microsoft.com/v1/admin/domains/delegatedTenantSettingOverrides?continuationToken={continuationToken}
    def list_domain_tenant_settings_overrides(self):
        """Returns list of tenant setting overrides that override at the domains

        Returns:
            list: The domains tenant settings overrides
        """
        url = "https://api.fabric.microsoft.com/v1/admin/domains/delegatedTenantSettingOverrides"

        items: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                           error_message = "Error listing domain tenant settings overrides",
                                           return_format="value", paging=True)

        return items
    
    def list_tenant_settings(self):
        """Get the tenant settings
        
        Returns:
            dict: The tenant settings
        """
        url = "https://api.fabric.microsoft.com/v1/admin/tenantsettings"

        response_json = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                             error_message = "Error getting tenant settings", paging= True, return_format="value")
        return response_json
    
    # GET https://api.fabric.microsoft.com/v1/admin/workspaces/delegatedTenantSettingOverrides?continuationToken={continuationToken}
    def list_workspace_tenant_settings_overrides(self):
        """Returns list of tenant setting overrides that override at the workspaces
        
        Returns:
            list: The workspaces tenant settings overrides
        """
        url = "https://api.fabric.microsoft.com/v1/admin/workspaces/delegatedTenantSettingOverrides"

        items: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                           error_message = "Error listing workspace tenant settings overrides",
                                           return_format="value", paging=True)

        return items
    
    # POST https://api.fabric.microsoft.com/v1/admin/capacities/{capacityId}/delegatedTenantSettingOverrides/{tenantSettingName}/update
    def update_capacity_tenant_setting_override(self, capacity_id, tenant_setting_name, enabled, delegate_to_workspace = None,
                                                enabled_security_groups = None, excluded_security_groups = None):
        """Update a tenant setting override at the capacity
        
        Args:
            capacity_id (str): The ID of the capacity
            tenant_setting_name (str): The name of the tenant setting
            enabled (bool): Whether the tenant setting is enabled
            delegate_to_workspace (bool): Indicates whether the tenant setting can be delegated to a workspace admin. False - Workspace admin cannot override the tenant setting. True - Workspace admin can override the tenant setting.
            enabled_security_groups (list): The list of enabled security groups
            excluded_security_groups (list): The list of excluded security groups
        Returns:
            dict: The overrides
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/capacities/{capacity_id}/delegatedTenantSettingOverrides/{tenant_setting_name}/update"
        body = {
            "enabled": enabled
        }
        if delegate_to_workspace:
            body["delegateToWorkspace"] = delegate_to_workspace
        if enabled_security_groups:
            body["enabledSecurityGroups"] = enabled_security_groups
        if excluded_security_groups:
            body["excludedSecurityGroups"] = excluded_security_groups

        response_json : dict = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                    error_message = "Error updating capacity tenant setting override", return_format="json")

        return response_json

    # POST https://api.fabric.microsoft.com/v1/admin/tenantsettings/{tenantSettingName}/update
    def update_tenant_setting(self, tenant_setting_name, enabled, delegate_to_capacity = None, delegate_to_domain = None,
                              delegate_to_workspace = None, enabled_security_groups = None, excluded_security_groups = None, properties = None):
        """Update a tenant setting

        Args:
            capacity_id (str): The ID of the capacity
            tenant_setting_name (str): The name of the tenant setting
            enabled (bool): Whether the tenant setting is enabled
            delegate_to_capacity (bool): Indicates whether the tenant setting can be delegated to a capacity admin. False - Capacity admin cannot override the tenant setting. True - Capacity admin can override the tenant setting.
            delegate_to_domain (bool): Indicates whether the tenant setting can be delegated to a domain admin. False - Domain admin cannot override the tenant setting. True - Domain admin can override the tenant setting.
            delegate_to_workspace (bool): Indicates whether the tenant setting can be delegated to a workspace admin. False - Workspace admin cannot override the tenant setting. True - Workspace admin can override the tenant setting.
            enabled_security_groups (list): The list of enabled security groups
            excluded_security_groups (list): The list of excluded security groups
        Returns:
            dict: The tenant settings
        """

        url = f"https://api.fabric.microsoft.com/v1/admin/tenantsettings/{tenant_setting_name}/update"

        body = {
            "enabled": enabled
        }
        if delegate_to_capacity:
            body["delegateToCapacity"] = delegate_to_capacity
        if delegate_to_domain:
            body["delegateToDomain"] = delegate_to_domain
        if delegate_to_workspace:
            body["delegateToWorkspace"] = delegate_to_workspace
        if enabled_security_groups:
            body["enabledSecurityGroups"] = enabled_security_groups
        if excluded_security_groups:
            body["excludedSecurityGroups"] = excluded_security_groups
        if properties:
            body["properties"] = properties

        response_json : dict = self.calling_routine(url = url, operation = "POST", body = body, response_codes = [200, 429],
                                                    error_message = "Error updating tenant setting", return_format="json")
        
        return response_json


    
    # Users APIs

    def list_access_entities(self, user_id, type = None):
        """Get the access entities for a user
        
        Args:
            user_id (str): The ID of the user
            type (str): The type of the access entity
        Returns:
            list: The list of access entities
        """

        url = f"https://api.fabric.microsoft.com/v1/admin/users/{user_id}/access"

        if type:
            url = f"{url}?type={type}"

        access_entities: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                                     error_message = "Error getting access entities", return_format="accessEntities", paging=True)
                
        return access_entities

    # Workspaces APIs
    
    def get_workspace(self, workspace_id):
        """Get a workspace by ID
        
        Args:
            workspace_id (str): The ID of the workspace
        Returns:
            Workspace: The Workspace object
        """
        from msfabricpysdkcore.admin_workspace import AdminWorkspace

        url = f"https://api.fabric.microsoft.com/v1/admin/workspaces/{workspace_id}"

        response_json: dict = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                                   error_message = "Error getting workspace", return_format="json")
    
        workspace = AdminWorkspace.from_dict(response_json, self)
        return workspace
    
    # GET https://api.fabric.microsoft.com/v1/admin/workspaces/discoverGitConnections
    def discover_git_connections(self):
        """Discover Git connections
        
        Returns:
            dict: The Git connections
        """
        url = "https://api.fabric.microsoft.com/v1/admin/workspaces/discoverGitConnections"

        response_json = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                             error_message = "Error discovering Git connections", return_format="value_json", paging=True)
        return response_json

    def list_workspace_access_details(self, workspace_id):
        """Get the access details of the workspace
        
        Args:
            workspace_id (str): The ID of the workspace
        Returns:
            dict: The access details of the workspace
        """
        url = f"https://api.fabric.microsoft.com/v1/admin/workspaces/{workspace_id}/users"

        response_json = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                             error_message = "Error getting workspace access details", return_format="json")
        return response_json
    

    def list_workspaces(self, capacity_id = None, name=None, state=None, type=None, continuationToken = None):
        """List all workspaces
        
        Args:
            capacity_id (str): The ID of the capacity
        Returns:
            list: List of Workspace objects
        """
        from msfabricpysdkcore.admin_workspace import AdminWorkspace

        url = "https://api.fabric.microsoft.com/v1/admin/workspaces"
        first_parameter = False
        if type:
            url = f"{url}?type={type}"
            first_parameter = True
        if capacity_id:
            if first_parameter:
                url = f"{url}&capacityId={capacity_id}"
            else:
                url = f"{url}?capacityId={capacity_id}"
                first_parameter = True

        if name:
            if first_parameter:
                url = f"{url}&name={name}"
            else:
                url = f"{url}?name={name}"
                first_parameter = True
        
        if state:
            if first_parameter:
                url = f"{url}&state={state}"
            else:
                url = f"{url}?state={state}"
                first_parameter = True

        workspaces: list = self.calling_routine(url = url, operation = "GET", response_codes = [200, 429],
                                                error_message = "Error listing workspaces", return_format="workspaces", paging=True)

        workspaces = [AdminWorkspace.from_dict(i, self) for i in workspaces]

        return workspaces
    
    def restore_workspace(self, workspace_id, new_workspace_admin_principal, new_workspace_name = None):
        """Restore a workspace
        
        Args:
            workspace_id (str): The ID of the workspace
        Returns:
            response: The response from the API
        """

        restore_request = {
            "newWorkspaceAdminPrincipal": new_workspace_admin_principal,
            "newWorkspaceName": new_workspace_name
        }
        url = f"https://api.fabric.microsoft.com/v1/admin/workspaces/{workspace_id}/restore"

        response_json = self.calling_routine(url = url, operation = "POST", body = restore_request,
                                             response_codes = [200, 429], error_message = "Error restoring workspace",
                                             return_format="response")

        return response_json



import json 

from msfabricpysdkcore.coreapi import FabricClientCore

class Folder:
    """Class to represent a folder in Microsoft Fabric"""

    def __init__(self, id, display_name, workspace_id, core_client: FabricClientCore, parent_folder_id) -> None:
        
        self.id = id
        self.display_name = display_name
        self.parent_folder_id = parent_folder_id
        self.workspace_id = workspace_id
        
        self.core_client = core_client

    def __str__(self) -> str:
        """Return a string representation of the workspace object"""
        dict_ = {
            'id': self.id,
            'display_name': self.display_name,
            'parent_folder_id': self.parent_folder_id,
            'workspace_id': self.workspace_id,
        }
        return json.dumps(dict_, indent=2)
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def from_dict(folder_dict, core_client):
        """Create Folder object from dictionary"""
        
        return Folder(id=folder_dict['id'], display_name=folder_dict['displayName'], workspace_id=folder_dict['workspaceId'],
                       parent_folder_id=folder_dict.get('parentFolderId', ""), core_client=core_client)

    def delete(self):
        """Delete the folder"""
        return self.core_client.delete_folder(workspace_id=self.workspace_id, folder_id=self.id)
    
    def get(self):
        """Get the folder"""
        returned_folder = self.core_client.get_folder(workspace_id=self.workspace_id, folder_id=self.id)
        self.display_name = returned_folder.display_name
        self.parent_folder_id = returned_folder.parent_folder_id
        self.workspace_id = returned_folder.workspace_id
        return returned_folder
    
    def move(self, target_folder_id = None):
        """Move a folder
        Args:
            target_folder_id (str): The ID of the target folder
        Returns:
            dict: The moved folder
        """
        moved_folder =  self.core_client.move_folder(folder_id=self.id, target_folder_id=target_folder_id)
        self.id = moved_folder.id
        self.display_name = moved_folder.display_name
        self.parent_folder_id = moved_folder.parent_folder_id
        self.workspace_id = moved_folder.workspace_id

        return moved_folder
    
    def update(self, display_name = None):
        """Update the folder"""
        updated_folder = self.core_client.update_folder(workspace_id=self.workspace_id, folder_id=self.id, display_name=display_name)
        self.display_name = updated_folder.display_name
        self.parent_folder_id = updated_folder.parent_folder_id
        self.workspace_id = updated_folder.workspace_id
        return updated_folder


import json 
from msfabricpysdkcore.coreapi import FabricClientCore

class OneLakeShortcut:
    """Class to represent a onelake shortcut in Microsoft Fabric"""

    def __init__(self, name, path, workspace_id, item_id, target,
                  core_client: FabricClientCore) -> None:
        
        self.name = name
        self.path = path
        self.target = target
        self.item_id = item_id
        self.workspace_id = workspace_id

        self.core_client = core_client

    def __str__(self) -> str:
        """Return a string representation of the workspace object"""
        dict_ = {
            'name': self.name,
            'path': self.path,
            'target': self.target,
            'item_id': self.item_id,
            'workspace_id': self.workspace_id,
        }
        return json.dumps(dict_, indent=2)
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def from_dict(short_dict, core_client):
        """Create OneLakeShortCut object from dictionary"""
        return OneLakeShortcut(name=short_dict['name'],
                               path=short_dict['path'], 
                               target=short_dict['target'], 
                               item_id=short_dict['itemId'], 
                               workspace_id=short_dict['workspaceId'], 
                               core_client=core_client)
    
    def delete(self):
        """Delete the shortcut"""
        return self.core_client.delete_shortcut(workspace_id=self.workspace_id, item_id=self.item_id, path=self.path,
                                                name=self.name)


import json 
from msfabricpysdkcore.coreapi import FabricClientCore

class JobInstance:
    """Class to represent a job instance in Microsoft Fabric"""

    def __init__(self, id, item_id, workspace_id, core_client: FabricClientCore, job_type, invoke_type, status, root_activity_id,
                 start_time_utc, end_time_utc, failureReason):

        self.id = id
        self.item_id = item_id
        self.workspace_id = workspace_id
        self.job_type = job_type
        self.invoke_type = invoke_type
        self.status = status
        self.root_activity_id = root_activity_id
        self.start_time_utc = start_time_utc
        self.end_time_utc = end_time_utc
        self.failureReason = failureReason
        
        self.core_client = core_client


    def __str__(self) -> str:
        """Return a string representation of the workspace object"""
        dict_ = {
            'id': self.id,
            'item_id': self.item_id,
            'workspace_id': self.workspace_id,
            'job_type': self.job_type,
            'invoke_type': self.invoke_type,
            'status': self.status,
            'root_activity_id': self.root_activity_id,
            'start_time_utc': self.start_time_utc,
            'end_time_utc': self.end_time_utc,
            'failureReason': self.failureReason
        }
        return json.dumps(dict_, indent=2)
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def from_dict(job_dict, core_client):
        """Create JobInstance object from dictionary"""
        return JobInstance(id=job_dict['id'], item_id=job_dict['itemId'], workspace_id=job_dict['workspaceId'],
                           job_type=job_dict['jobType'], invoke_type=job_dict['invokeType'], status=job_dict['status'],
                           root_activity_id=job_dict['rootActivityId'], start_time_utc=job_dict['startTimeUtc'],
                           end_time_utc=job_dict['endTimeUtc'], failureReason=job_dict['failureReason'], core_client=core_client)
    
    def cancel(self):
        """Cancel the job instance"""
        return self.core_client.cancel_item_job_instance(workspace_id=self.workspace_id,
                                                         item_id=self.item_id,
                                                         job_instance_id=self.id)
    



from warnings import warn

from msfabricpysdkcore.client import FabricClient


class FabricAzureClient(FabricClient):

    def __init__(self, tenant_id=None, client_id=None, client_secret=None, 
                 username = None, password = None, silent=None) -> None:
        super().__init__(scope = "https://management.azure.com/.default",
                         tenant_id = tenant_id,
                         client_id = client_id,
                         client_secret = client_secret,
                         username = username,
                         password = password)

        if silent is not None:
            warn("The 'silent' parameter is deprecated and will be removed in a future version.", DeprecationWarning, stacklevel=2)


    def check_name_availability(self, subscription_id, location, name, type = "Microsoft.Fabric/capacities"):
        """Check name availability
        Args:
            subscription_id (str): The subscription ID
            location (str): The location
            name (str): The name
            type (str): The type
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Fabric/locations/{location}/checkNameAvailability?api-version=2023-11-01"

        body = {
            "name": name,
            "type": type
        }

        response = self.calling_routine(url=url, operation="POST", body=body, response_codes=[200], return_format="json", error_message="Failed to check name availability")
        return response

    def create_or_update_capacity(self, subscription_id, resource_group_name, capacity_name, location, properties_administration, sku, tags = None):
        """Create or update capacity
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
            capacity_name (str): The capacity name
            location (str): The location
            properties_administration (dict): The administration properties
            sku (dict): The sku
            tags (dict): The tags
        Returns:
            FabricAzureCapacity: The capacity
        """
        from msfabricpysdkcore.fabric_azure_capacity import FabricAzureCapacity

        if sku and "name" in sku:
            sku = sku["name"]

        body = {
            "location": location,
            "properties": {
                "administration": properties_administration
            },
            "sku": {"name": sku,
                    "tier": "Fabric"}
        }

        if tags is not None:
            body["tags"] = tags
        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities/{capacity_name}?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="PUT", body=body, response_codes=[200, 201], return_format="json", error_message="Failed to create or update capacity")
        response["subscription_id"] = subscription_id
        response["resource_group_name"] = resource_group_name

        return FabricAzureCapacity.from_dict(response, self)


    def delete_capacity(self, subscription_id, resource_group_name, capacity_name):
        """Delete capacity
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
            capacity_name (str): The capacity name
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities/{capacity_name}?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="DELETE", response_codes=[202], return_format="response", error_message="Failed to delete capacity")
        return response


    def get_capacity(self, subscription_id, resource_group_name, capacity_name):
        """Get capacity
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
            capacity_name (str): The capacity name
        Returns:
            FabricAzureCapacity: The capacity
        """
        from msfabricpysdkcore.fabric_azure_capacity import FabricAzureCapacity

        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities/{capacity_name}?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="GET", response_codes=[200], return_format="json", error_message="Failed to get capacity")
        response["subscription_id"] = subscription_id
        response["resource_group_name"] = resource_group_name

        return FabricAzureCapacity.from_dict(response, self)

    def list_by_resource_group(self, subscription_id, resource_group_name):
        """List capacities by resource group
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="GET", response_codes=[200], return_format="value_json", error_message="Failed to list capacities by resource group")
        return response


    def list_by_subscription(self, subscription_id):
        """List capacities by subscription
        Args:
            subscription_id (str): The subscription ID
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Fabric/capacities?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="GET", response_codes=[200], return_format="value_json", error_message="Failed to list capacities by subscription")
        return response

    def list_skus(self, subscription_id):
        """List skus
        Args:
            subscription_id (str): The subscription ID
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Fabric/skus?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="GET", response_codes=[200], return_format="value_json", error_message="Failed to list skus")
        return response
    
    def list_skus_for_capacity(self, subscription_id, resource_group_name, capacity_name):
        """List skus for capacity
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
            capacity_name (str): The capacity name
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities/{capacity_name}/skus?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="GET", response_codes=[200], return_format="value_json", error_message="Failed to list skus for capacity")
        return response

    def resume_capacity(self, subscription_id, resource_group_name, capacity_name):
        """Resume capacity
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
            capacity_name (str): The capacity name
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities/{capacity_name}/resume?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="POST", response_codes=[202], return_format="response", error_message="Failed to resume capacity")
        return response

    def suspend_capacity(self, subscription_id, resource_group_name, capacity_name):
        """Suspend capacity
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
            capacity_name (str): The capacity name
        Returns:
            dict: The response
        """

        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities/{capacity_name}/suspend?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="POST", response_codes=[202], return_format="response", error_message="Failed to suspend capacity")
        return response
    
    def update_capacity(self, subscription_id, resource_group_name, capacity_name, properties_administration = None, sku = None, tags = None):
        """Update capacity
        Args:
            subscription_id (str): The subscription ID
            resource_group_name (str): The resource group name
            capacity_name (str): The capacity name
            body (dict): The body of the request
        Returns:
            FabricAzureCapacity: The capacity
        """
        from msfabricpysdkcore.fabric_azure_capacity import FabricAzureCapacity

        body = {}
        if sku and "name" in sku:
            sku = sku["name"]

        if properties_administration is not None:
            body["properties"] = {}
            body["properties"]["administration"] = properties_administration

        if sku is not None:
            body["sku"] = {"name": sku,
                           "tier": "Fabric"}
        
        if tags is not None:
            body["tags"] = tags

        url = f"https://management.azure.com/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Fabric/capacities/{capacity_name}?api-version=2023-11-01"

        response = self.calling_routine(url=url, operation="PATCH", body=body, response_codes=[200, 202],
                                        return_format="json", error_message="Failed to update capacity")
        response["subscription_id"] = subscription_id
        response["resource_group_name"] = resource_group_name

        return FabricAzureCapacity.from_dict(response, self)



from msfabricpysdkcore.item import Item

class ApacheAirflowJob(Item):
    """Class to represent a ApacheAirflowJob in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return ApacheAirflowJob(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class CopyJob(Item):
    """Class to represent a copy job in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return CopyJob(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class Dataflow(Item):
    """Class to represent a dataflow in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return Dataflow(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
class DataPipeline(Item):
    """Class to represent a spark job definition in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return DataPipeline(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
        
    def run_on_demand_item_job(self, execution_data=None):
        return self.core_client.run_on_demand_item_job(workspace_id=self.workspace_id, item_id=self.id, job_type="Pipeline", execution_data=execution_data)

class DigitalTwinBuilder(Item):
    """Class to represent a DigitalTwinBuilder in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return DigitalTwinBuilder(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class DigitalTwinBuilderFlow(Item):
    """Class to represent a DigitalTwinBuilderFlow in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return DigitalTwinBuilderFlow(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class MirroredAzureDatabricksCatalog(Item):
    """Class to represent a mirrored Azure Databricks catalog in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return MirroredAzureDatabricksCatalog(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def refresh_catalog_metadata(self):
        """Method to refresh the metadata of the mirrored Azure Databricks catalog"""
        return self.core_client.refresh_mirrored_azure_databricks_catalog_metadata(self.workspace_id, self.id)

class Eventhouse(Item):
    """Class to represent a eventhouse in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, description=""):
        super().__init__(id = id, display_name=display_name, type=type, 
                         workspace_id=workspace_id, core_client=core_client, properties=properties, 
                         description=description)
        
    def from_dict(item_dict, core_client):
        if "displayName" not in item_dict:
            item_dict["displayName"] = item_dict["display_name"]
        if "workspaceId" not in item_dict:
            item_dict["workspaceId"] = item_dict["workspace_id"]

        return Eventhouse(id=item_dict['id'], display_name=item_dict['displayName'], 
                          type=item_dict['type'], workspace_id=item_dict['workspaceId'],
                          properties=item_dict.get('properties', None),
                          description=item_dict.get('description', ""), core_client=core_client)

    def get_definition(self, type=None, format=None):
        """Method to get the definition of the eventhouse"""
        return super().get_definition(type="eventhouses", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the eventhouse"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="eventhouses")

    def create_kql_database(self, display_name = None, description= None):
        from msfabricpysdkcore.coreapi import FabricClientCore
        """Method to create a kql database in the eventhouse"""
        creation_payload = {"databaseType" : "ReadWrite",
                            "parentEventhouseItemId" : self.id}
        
        return self.core_client.create_kql_database(self.workspace_id, display_name=display_name, description=description, creation_payload=creation_payload)

class SparkJobDefinition(Item):
    """Class to represent a spark job definition in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return SparkJobDefinition(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def get_definition(self, format=None):
        return super().get_definition(type="sparkJobDefinitions", format=format)
    
    def update_definition(self, definition):
        return self.core_client.update_spark_job_definition_definition(self.workspace_id, self.id, definition)
    
    def run_on_demand_spark_job_definition(self, job_type = "sparkjob"):
        return self.core_client.run_on_demand_spark_job_definition(workspace_id=self.workspace_id, spark_job_definition_id=self.id, job_type=job_type)

    def list_livy_sessions(self):
        """List all livy sessions in the spark job definition"""
        return self.core_client.list_spark_job_definition_livy_sessions(self.workspace_id, self.id)

    def get_livy_session(self, livy_id):
        """Get a livy session in the spark job definition"""
        return self.core_client.get_spark_job_definition_livy_session(self.workspace_id, self.id, livy_id)

class Warehouse(Item):
    """Class to represent a warehouse in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return Warehouse(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
class WarehouseSnapshot(Item):
    """Class to represent a warehouse snapshot in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return WarehouseSnapshot(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class KQLDashboard(Item):
    """Class to represent a kql dashboard in Microsoft Fabric"""

    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return KQLDashboard(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def get_definition(self, format=None):
        """Method to get the definition of the kql dashboard"""
        return super().get_definition(type="kqlDashboards", format=format)

    def update_definition(self, definition):
        """Method to update the definition of the kql dashboard"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="kqlDashboards")

class KQLDatabase(Item):
    """Class to represent a kql database in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return KQLDatabase(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

    def get_definition(self, type=None, format=None):
        """Method to get the definition of the kql database"""
        return super().get_definition(type="kqlDatabases", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the kql database"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="kqlDatabases")

class KQLQueryset(Item):
    """Class to represent a kql database in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return KQLQueryset(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def get_definition(self, format=None):
        """Method to get the definition of the kql queryset"""
        return super().get_definition(type="kqlQuerysets", format=format)
    
    def update_definition(self, definition, update_metadata=None):
        """Method to update the definition of the kql queryset"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="kqlQuerysets",
                                                       update_metadata=update_metadata)
    

class GraphQLApi(Item):
    """Class to represent a graphql api in Microsoft Fabric"""

    def __init__(self, id, display_name, type, workspace_id, core_client, properties=None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)
    
    def from_dict(item_dict, core_client):
        return GraphQLApi(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)



class MirroredDatabase(Item):
    """Class to represent a mirrored database in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)
    
    def from_dict(item_dict, core_client):
        return MirroredDatabase(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def get_definition(self): 
        """Method to get the definition of the mirrored database"""
        return super().get_definition(type="mirroredDatabases")
    
    def update_definition(self, definition):
        """Method to update the definition of the mirrored database"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="mirroredDatabases")

    def get_mirrored_database_definition(self, mirrored_database_id):
        return self.core_client.get_mirrored_database_definition(workspace_id=self.id,
                                                                 mirrored_database_id=mirrored_database_id)
    def get_mirroring_status(self):
        return self.core_client.get_mirroring_status(workspace_id=self.workspace_id, mirrored_database_id=self.id)

    def get_tables_mirroring_status(self):
        return self.core_client.get_tables_mirroring_status(workspace_id=self.workspace_id, mirrored_database_id=self.id)

    def start_mirroring(self):
        return self.core_client.start_mirroring(workspace_id=self.workspace_id, mirrored_database_id=self.id)

    def stop_mirroring(self):
        return self.core_client.stop_mirroring(workspace_id=self.workspace_id, mirrored_database_id=self.id)

class MLExperiment(Item):
    """Class to represent a ml experiment in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)
    
    def from_dict(item_dict, core_client):
        return MLExperiment(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class MLModel(Item):
    """Class to represent a ml model in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return MLModel(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class MountedDataFactory(Item):
    """Class to represent a mounted data factory in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)
    
    def from_dict(item_dict, core_client):
        return MountedDataFactory(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

    def get_definition(self, type=None, format=None):
        """Method to get the definition of the mountedDataFactory"""
        return super().get_definition(type="mountedDataFactories", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the mountedDataFactory"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="mountedDataFactories")

class Notebook(Item):
    """Class to represent a notebook in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return Notebook(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def get_definition(self, format=None):
        """Method to get the definition of the notebook"""
        return super().get_definition(type="notebooks", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the notebook"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="notebooks")
    
    def list_livy_sessions(self):
        """List all livy sessions in the notebook"""
        return self.core_client.list_notebook_livy_sessions(self.workspace_id, self.id)

    def get_livy_session(self, livy_id):
        """Get a livy session in the notebook"""
        return self.core_client.get_notebook_livy_session(self.workspace_id, self.id, livy_id)

class Reflex(Item):
    """Class to represent a reflex in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return Reflex(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

    def get_definition(self, type=None, format=None):
        """Method to get the definition of the reflex"""
        return super().get_definition(type="reflexes", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the reflex"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="reflexes")
    
class Report(Item):
    """Class to represent a report in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return Report(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def get_definition(self, type=None, format=None):
        """Method to get the definition of the report"""
        return super().get_definition(type="reports", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the report"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="reports")

class SemanticModel(Item):
    """Class to represent a semantic model in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return SemanticModel(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)
    
    def get_definition(self, format=None):
        """Method to get the definition of the semantic model"""
        return super().get_definition(type="semanticModels", format=format)
    
    def update_definition(self, definition):
        """Method to update the definition of the semantic model"""
        return self.core_client.update_item_definition(self.workspace_id, self.id, definition, type="semanticModels")

class SQLDatabase(Item):
    """Class to represent a sql database in Microsoft Fabric"""
     
    def __init__(self, id, display_name, type, workspace_id, core_client, properties = None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return SQLDatabase(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)

class VariableLibrary(Item):
    """Class to represent a variable library in Microsoft Fabric"""

    def __init__(self, id, display_name, type, workspace_id, core_client, properties=None, definition=None, description=""):
        super().__init__(id, display_name, type, workspace_id, core_client, properties, definition, description)

    def from_dict(item_dict, core_client):
        return VariableLibrary(id=item_dict['id'], display_name=item_dict['displayName'], type=item_dict['type'], workspace_id=item_dict['workspaceId'],
            properties=item_dict.get('properties', None),
            definition=item_dict.get('definition', None), description=item_dict.get('description', ""), core_client=core_client)


import json
from msfabricpysdkcore.adminapi import FabricClientAdmin


class AdminItem:
    """Class to represent a item in Microsoft Fabric"""

    def __init__(self, id, type, name, workspace_id, state, description, last_updated_date, capacity_id, creator_principal, admin_client: FabricClientAdmin) -> None:
        """Constructor for the Item class
        
        Args:
            id (str): The ID of the item
            type (str): The type of the item
            name (str): The name of the item
            workspace_id (str): The ID of the workspace to which the item belongs
            state (str): The state of the item
            description (str): The description of the item
            last_updated_date (str): The date when the item was last updated
            capacity_id (str): The
            creator_principal (str): The principal who created the item
            admin_client (FabricClientAdmin): The FabricClientAdmin object
        """

        self.id = id
        self.type = type
        self.name = name
        self.workspace_id = workspace_id
        self.state = state
        self.description = description
        self.last_updated_date = last_updated_date
        self.capacity_id = capacity_id
        self.creator_principal = creator_principal
        self.admin_client = admin_client


    def __str__(self) -> str:
        """Return a string representation of the workspace object
        
        Returns:
            str: The string representation of the workspace object
        """
        dict_ = {
            'id': self.id,
            'type': self.type,
            'name': self.name,
            'workspace_id': self.workspace_id,
            'state': self.state,
            'description': self.description,
            'last_updated_date': self.last_updated_date,
            'capacity_id': self.capacity_id,
            'creator_principal': self.creator_principal
        }
        return json.dumps(dict_, indent=2)

    
    def __repr__(self) -> str:
        return self.__str__()
    
    def from_dict(item_dict, admin_client):
        """Create Item object from dictionary
        
        Args:
            item_dict (dict): The dictionary containing the item details
            admin_client (FabricClientAdmin): The FabricClientAdmin object
        Returns:
            AdminItem: The AdminItem object
        """
        return AdminItem(
            id = item_dict['id'],
            type = item_dict['type'],
            name = item_dict.get('name', None),
            workspace_id = item_dict['workspaceId'],
            state = item_dict['state'],
            description = item_dict.get('description', None),
            last_updated_date = item_dict['lastUpdatedDate'],
            capacity_id = item_dict['capacityId'],
            creator_principal = item_dict.get('creatorPrincipal', None),
            admin_client = admin_client
        )
            
    def list_item_access_details(self, type=None):
        """Get the access details of the item
        
        Returns:
            dict: The access details of the item"""
        
        return self.admin_client.list_item_access_details(self.workspace_id, self.id, type)